# = Formatters
#
# Formatters are responsible for formatting log event data.
# An Outputter owning a Formatter will invoke the 
# Log4r::Formatter#format method prior to writing.
#
# === Available Formatters
#
# * Log4r::BasicFormatter -  default
# * Log4r::PatternFormatter - most flexible. See log4r/patternformatter.rb
# * Log4r::SimpleFormatter - like BasicFormatter for Strings only (low noise)
# * Log4r::ObjectFormatter - for inspecting objects
# * Log4r::NullFormatter - twirls on its feet and does nothing
#
# === XML Configuration
#
# Specify the Formatter and its class (as +type+) under an 
# <tt><outputter></tt> directive:
#
#   <outputter name="someout" type="sometype">
#     <formatter type="Log4r::BasicFormatter"/>
#   </outputter>
#
# As explained in log4r/configurator.rb, the hash arguments you would normally
# pass to +new+ are specified as <i>XML parameters</i>. Only PatternFormatter
# has any of these.
#
# === custom Formatting
#
# Building a custom Formatter is extremely easy. Just define a class
# that extends Formatter and override the Formatter#format method.
# Then give it to any interested Outputters. 
#
# If you're interested in setting up your custom formatters in XML,
# please take a look at log4r/configurator.rb.
#
# === Data Available
#
# The following data is passed to the format method:
#
# level::    The integer level of the log event.
# logger::   The logger which was invoked.
# tracer::   The execution stack at the log event, nil if logger not tracing.
# data::     The object passed to the log method.
# 
# Within the logger is some extra data about the logger. Please see the
# section <b>Inheritance</b> in log4r/logger.rb
#
# Version:: $Id: formatter.rb,v 1.11 2002/01/16 13:13:57 cepheus Exp $

require "singleton"

require "log4r/base"

module Log4r

  # Formatter is an abstract class
  class Formatter
    # Raises an exception if called directly.
    def initialize
      if type == Formatter
        raise TypeError, "Formatter is an abstract class!"
      end
    end

    # Define this method in a subclass to format data.
    #
    # [level]    The integer level of the log event. Use LNAMES[level]
    #            to get the actual level name.
    # [logger]   The logger that was invoked. Contains useful data.
    #            Please see Inheritance in log4r/logger.rb for details.
    # [tracer]   The execution stack returned by <tt>caller</tt> at the
    #            log event. It is nil if the invoked Logger's trace is false.
    # [data]     The object that was passed into the logging method.

    def format(level, logger, tracer, data)
    end
  end

  # A null formatter, or toy if you prefer. Good for unit testing.
  class NullFormatter < Formatter
    include Singleton
  end
  
  # SimpleFormatter produces output like this:
  # 
  #   WARN loggername> Danger, Will Robinson, danger!
  #
  # Does not write traces and does not inspect objects.
  
  class SimpleFormatter < Formatter
    @@simpleformat = "%*4$s %s> %s\n"
    def format(level, logger, tracer, data)
      sprintf @@simpleformat, LNAMES[level], logger.name, data, MaxLevelLength
    end
  end

  # BasicFormatter produces output like this:
  # 
  #   WARN loggername: I dropped my Wookie!
  #   
  # Or like this if trace is on:
  # 
  #   WARN loggername(file.rb at 12): Hot potato!
  #   
  # Also, it will pretty-print any Exception it gets and
  # <tt>inspect</tt> everything else.

  class BasicFormatter < SimpleFormatter
    @@formatspec = "%*3$s %s"
    def format(level, logger, tracer, data)
      buff = sprintf(@@formatspec, LNAMES[level], logger.name, MaxLevelLength)
      buff += (tracer.nil? ? "" : "(#{tracer[0]})") + ": "
      buff += format_object(data) + "\n"
      buff
    end

    # Formats data according to its class:
    #
    # String::     Prints it out as normal.
    # Exception::  Produces output similar to command-line exceptions.
    # Object::     Prints the type of object, then the output of
    #              <tt>inspect</tt>. An example -- Array: [1, 2, 3]

    def format_object(obj)
      if obj.kind_of? Exception
        return "Caught #{obj.type}: #{obj.message}" + \
               "\n   " + obj.backtrace.join("\n   ")
      elsif obj.kind_of? String
        return obj
      else # inspect the object
        return "#{obj.type}: #{obj.inspect}"
      end
    end
  end

  # Formats objects the same way irb does:
  #
  #   loggername:foo.rb in 12> 
  #   [1, 3, 4]
  #   loggername:foo.rb in 13> 
  #   {1=>"1"}
  #
  # Strings don't get inspected. just printed. The trace is optional.

  class ObjectFormatter < Formatter
    def format(level, logger, tracer, data)
      buff = logger.name
      buff += (tracer.nil? ? "" : ":#{tracer[0]}") + ">\n"
      buff += (data.kind_of?(String) ?  data : data.inspect) + "\n"
      buff
    end
  end
  
  # Outputters that don't define a Formatter will get this, which
  # is currently BasicFormatter
  class DefaultFormatter < BasicFormatter
  end
  
end
