# :include: rdoc/logger
#
# == Other Info
# 
# Version:: $Id: logger.rb,v 1.18 2002/01/16 06:28:07 cepheus Exp $
# Author:: Leon Torres <leon@ugcs.caltech.edu>

require "thread"

require "log4r/outputter"
require "log4r/repository"
require "log4r/loggerfactory"

module Log4r
  
  class Logger
    attr_reader :name, :fullname, :path, :level, :parent
    attr_reader :additive, :trace, :outputters

    # Logger requires a name. The last 3 parameters are:
    #
    # level::     Do I have a level? (Otherwise, I'll inherit my parent's)
    # additive::  Am I additive?
    # trace::     Do I record the execution trace?

    def initialize(_fullname, _level=nil, _additive=true, _trace=false)
      # validation
      raise ArgumentError, "Logger must have a name", caller if _fullname.nil?
      Log4rTools.validate_level(_level) unless _level.nil?
      validate_name(_fullname)
      
      # create the logger
      @fullname = _fullname
      @outputters = []
      @additive = _additive
      deal_with_inheritance(_level)
      LoggerFactory.define_methods(self)
      self.trace = _trace
      Repository[@fullname] = self
    end

    def validate_name(_fullname)
      parts = _fullname.split Log4rConfig::LoggerPathDelimiter
      for part in parts
        raise ArgumentError, "Malformed path", caller[1..-1] if part.empty?
      end
    end
    private :validate_name

    # Parses name for location in heiarchy, sets the parent, and
    # deals with level inheritance

    def deal_with_inheritance(_level)
      mypath = @fullname.split Log4rConfig::LoggerPathDelimiter
      @name = mypath.pop
      if mypath.empty? # then root is my daddy
        @path = ""
        # This is one of the guarantees that RootLogger gets created
        @parent = Logger.root
      else
        @path = mypath.join(Log4rConfig::LoggerPathDelimiter)
        @parent = Repository.find_ancestor(@path)
        @parent = Logger.root if @parent.nil?
      end
      # inherit the level if no level defined
      if _level.nil? then @level = @parent.level
      else @level = _level end
      Repository.reassign_any_children(self)
    end
    private :deal_with_inheritance

    # Set the logger level dynamically. Does not affect children.
    def level=(_level)
      Log4rTools.validate_level(_level)
      @level = _level
      LoggerFactory.define_methods(self)
      @level
    end

    # Set the additivity of the logger dynamically. True or false.
    def additive=(_additive)
      @additive = _additive
      LoggerFactory.define_methods(self)
      @additive
    end

    # Set whether the logger traces. Can be set dynamically. True or false.
    def trace=(_trace)
      case _trace
      when "true", true then @trace = true
      else @trace = false
      end
      LoggerFactory.define_methods(self)
      @trace
    end

    # Please don't reset the parent
    def parent=(parent)
      @parent = parent
    end
    
    # Set the Outputters dynamically by name or reference.
    def outputters=(_outputters)
      @outputters.clear
      add(*_outputters)
    end  

    # Add outputters by name or by reference. May be done dynamically. 
    def add(*_outputters)
      for thing in _outputters
        o = (thing.kind_of?(Outputter) ? thing : Outputter[thing])
        # some basic validation
        if not o.kind_of?(Outputter)
          raise TypeError, "Expected kind of Outputter, got #{o.type}", caller
        elsif o.nil?
          raise TypeError, "Couldn't find Outputter '#{thing}'", caller
        end
        @outputters.push o
      end
      @outputters
    end

    # Remove outputters from this logger by name.
    def remove(*_outputters)
      for name in _outputters
        o = Outputter[name]
        @outputters.delete o
      end
    end

    def is_root?; false end

    def ==(other)
      return true if id == other.id
    end

    class << self
      # Returns the root logger. Identical to Logger.global
      def root; return RootLogger.instance end
      # Returns the root logger. Identical to Logger.root
      def global; return root end
      
      # Get a logger with a fullname from the repository. To save us some pain,
      # it raises NameError if the logger is not found.

      def [](_fullname)
        # forces creation of RootLogger if it doesn't exist yet.
        return RootLogger.instance if _fullname == 'root' or _fullname == 'global'
        logger = Repository[_fullname]
        if logger.nil?
          raise NameError, "Logger '#{_fullname}' not found.", caller
        end
        return logger
      end

      # Yields fullname and logger for every logger in the system.
      def each
        for fullname, logger in Repository.instance.loggers
          yield fullname, logger
        end
      end

    end
  end

  # RootLogger should be retrieved with Logger.root or Logger.global.
  # It's supposed to be transparent.
  #--
  # We must guarantee the creation of RootLogger before any other Logger
  # or Outputter gets their logging methods defined. There are two 
  # guarantees in the code:
  #
  # * Logger#deal_with_inheritance - calls RootLogger.instance when
  #   a new Logger is created without a parent. Parents must exist, therefore
  #   RootLogger is forced to be created.
  #
  # * OutputterFactory.create_methods - Calls Logger.root first. So if
  #   an Outputter is created, RootLogger is also created.
  #
  # When RootLogger is created, it calls
  # Log4r.define_levels(*Log4rConfig::LogLevels). This ensures that the 
  # default levels are loaded if no custom ones are.

  class RootLogger < Logger
    include Singleton

    def initialize
      Log4r.define_levels(*Log4rConfig::LogLevels) # ensure levels are loaded
      @level = ALL
      @outputters = []
      Repository['root'] = self
      Repository['global'] = self
      LoggerFactory.undefine_methods(self)
    end

    def is_root?; true end

    # Set the global level. Any loggers defined thereafter will
    # not log below the global level regardless of their levels.

    def level=(alevel); @level = alevel end

    # Does nothing
    def outputters=(foo); end
    # Does nothing
    def trace=(foo); end
    # Does nothing
    def additive=(foo); end
    # Does nothing
    def add(*foo); end
    # Does nothing
    def remove(*foo); end
  end
end
