# = Outputters
# 
# An Outputter is a logging destination. It is given an IO object to write 
# to, a Formatter to call, and, optionally, levels to write at.
#
# Outputters invoke print then flush on the wrapped IO object.
# If the IO chokes, the Outputter will close the IO and set its
# level to <tt>OFF</tt>.
#
# ==== Outputter Levels
# 
# Outputters have their own level thresholds that default to <tt>root</tt> 
# level. They will not write any log events with a rank less than their
# threshold.
#
# Alternatively, an Outputter can be told to log specific levels only:
#
#   o = StdoutOutputter.new 'console'
#   o.only_at DEBUG, FATAL         # only DEBUG and FATAL get written 
#
# ==== Outputter Repository
#
# When outputters are created, they store themselves in an Outputter
# repository similar to the Logger repository. Outputters require names
# so that they may be retrieved with ease.
# 
#   StdoutOutputter.new 'console' => Create 'console' outputter
#   Outputter['console']          => Get it back from the stash.
#
# ==== Default Outputters
#
# Two outputters named 'stdout' and 'stderr' are created automatically at
# the root level. You can add these to a Logger as a shortcut.
#
#   Outputter['stdout'] => 'stdout'
#   Outputter['stderr'] => 'stderr'
#   Outputter.stdout    => 'stdout'
#   Outputter.stderr    => 'stderr'
#
# ==== Configuring
#
# Outputters must have names and receive hash arguments. The parameter name
# for the hash args can be either a symbol or a string. All defined outputters
# accept <tt>:level</tt> and <tt>:formatter</tt> arguments. For arguments 
# specific to a convenience Outputter, please look at the class description.
#
# You can change the level threshold, levels logging (only_at), and formatter
# dynamically using the <tt>=</tt> methods.
#
# Here are some ways to set up an Outputter:
# 
#   Outputter.new ExoticIO.new 'exotic', 'level' => WARN, 
#                 :formatter => MyFormatter.new
#   # an equivalent way:
#   o = Outputter.new ExoticIO.new 'exotic'
#   o.level = WARN
#   o.formatter = MyFormatter         # we can specify just the class
#
# ==== Convenience Outputters
#
# * Log4r::StdoutOutputter - writes to $stdout
# * Log4r::StderrOutputter - writes to $stderr
# * Log4r::FileOutputter - opens and writes to a file
#
# ==== XML Configuration
#
# Specify outputters as children of <tt><log4r_config></tt>:
#
#   <log4r_config>
#     <outputter name="myout" type="Log4r::StdoutOutputter">
#       <only_at>DEBUG, INFO</only_at>
#     </outputter>
#     <outputter name="file" level="WARN">
#       <type>FileOutputter</type>
#       <filename>#{logpath}/file.log</filename>
#       <trunc>false</trunc>
#     </outputter>
#     ...
#
# As explained in log4r/configurator.rb, the hash arguments you would normally
# pass to <tt>new</tt> are specified as <i>XML parameters</i>.
#   
# == Other Info
#
# Version:: $Id: outputter.rb,v 1.20 2002/01/16 06:42:38 cepheus Exp $
# Author:: Leon Torres <leon@ugcs.caltech.edu>

require "thread"

require "log4r/outputterfactory"
require "log4r/formatter"

module Log4r

  class Outputter
    attr_reader :level, :formatter
    @@outputters = Hash.new

    # An Outputter needs an IO object to write to. The hash arguments
    # are as follows:
    # 
    # [<tt>:level</tt>]       Logger level. Optional, defaults to root level
    # [<tt>:formatter</tt>]   A Formatter. Defaults to DefaultFormatter

    def initialize(_name, _out, hash=nil)
      if _out.nil? or _name.nil?
        raise ArgumentError, "Bad arguments. Name and IO expected.", caller
      elsif !_out.kind_of?(IO)
        raise TypeError, "IO object expected in second argumenT", caller
      end
      validate_hash(hash)
      @out = _out
      @name = _name
      @@outputters[@name] = self
    end

    def closed?
      @out.closed?
    end

    # Close the IO and sets level to OFF
    def close
      @out.close
      @level = OFF
      OutputterFactory.create_methods(self)
    end

    # dynamically change the level
    def level=(_level)
      Log4rTools.validate_level(_level)
      @level = _level
      OutputterFactory.create_methods(self)
    end

    # Set the levels to log. All others will be ignored
    def only_at(*levels)
      raise ArgumentError, "Gimme some levels!", caller if levels.empty?
      raise ArgumentError, "Can't log only_at ALL", caller if levels.include? ALL
      levels.each {|level| Log4rTools.validate_level(level)}
      @level = levels.sort.first
      OutputterFactory.create_methods self, levels
    end

    # Dynamically change the formatter. You can just specify a Class
    # object and the formatter will invoke +new+ or +instance+
    # on it as appropriate.
 
    def formatter=(_formatter)
      if _formatter.kind_of?(Formatter)
        @formatter = _formatter
      elsif _formatter.kind_of?(Class) and _formatter <= Formatter
        if _formatter.respond_to? :instance
          @formatter = _formatter.instance
        else
          @formatter = _formatter.new
        end
      else
        raise TypeError, "Argument was not a Formatter!", caller
      end
    end

    # Validates the common hash arguments. For now, that would be
    # +:level+, +:formatter+ and the string equivalents
    def validate_hash(hash)
      @mutex = Mutex.new
      # default to root level and DefaultFormatter
      if hash.nil?
        self.level = Logger.root.level
        @formatter = DefaultFormatter.new
        return
      end
      self.level = (hash[:level] or hash['level'] or Logger.root.level)
      self.formatter = (hash[:formatter] or hash['formatter'] or DefaultFormatter.new)
    end
    protected :validate_hash

    class << self
      # Retrieve an outputter.
      def [](name)
        out = @@outputters[name]
        if out.nil?
          return case name
          when 'stdout' then StdoutOutputter.new 'stdout'
          when 'stderr' then StderrOutputter.new 'stderr'
          else nil end
        end          
        out
      end
      def stdout; Outputter['stdout'] end
      def stderr; Outputter['stderr'] end
      # Set an outputter.
      def []=(name, outputter)
        @@outputters[name] = outputter
      end
      # Yields each outputter's name and reference.
      def each
        @@outputters.each {|name, outputter| yield name, outputter}
      end
    end

    #######
    private
    #######

    # perform the write
    def write(data)
      begin
        @out.print data
        @out.flush
      rescue NameError => ne
        # out was *somehow* turned to non-IO object. this is really bad
        raise NameError, 
              "Fatal: IO in Outputter changed to non-IO object! " + \
              "IO became #{ne.type}", caller[1..-2]
      rescue IOError
        # we want to recover from this instead of crash
        close
      end
    end
    def synch; @mutex.synchronize { yield } end
  end
end
